package org.osm2world.core.target.povray;

import java.io.File;
import java.io.IOException;
import java.io.PrintStream;

import org.osm2world.core.heightmap.data.CellularTerrainElevation;
import org.osm2world.core.map_data.data.MapData;
import org.osm2world.core.map_data.data.MapElement;
import org.osm2world.core.math.VectorXYZ;
import org.osm2world.core.math.VectorXZ;
import org.osm2world.core.target.RenderableToAllTargets;
import org.osm2world.core.target.common.rendering.Camera;
import org.osm2world.core.target.common.rendering.Projection;
import org.osm2world.core.terrain.data.Terrain;
import org.osm2world.core.world.data.WorldObject;
import org.osm2world.core.world.modules.common.Materials;

import static org.osm2world.core.util.FaultTolerantIterationUtil.*;

/**
 * utility class for creating a POVRay file
 */
public final class POVRayWriter {

	/** prevents instantiation */
	private POVRayWriter() { }
	
	public static final void writePOVInstructionFile(File file, MapData grid, 
			CellularTerrainElevation eleData, Terrain terrain,
			Camera camera, Projection projection)
			throws IOException {
		
		if (!file.exists()) {
			file.createNewFile();
		}
		PrintStream printStream = new PrintStream(file);
		
		writePOVInstructionStringToStream(printStream,
				grid, eleData, terrain, camera, projection);
		
	}

	private static final void writePOVInstructionStringToStream(
			PrintStream stream,
			MapData grid, 
			CellularTerrainElevation eleData, Terrain terrain,
			Camera camera, Projection projection) {
				
		POVRayTarget target = new POVRayTarget(stream);

		addRenderParameters(target);
		
		target.append("\n#include \"textures.inc\"\n#include \"colors.inc\"\n");
		target.append("#include \"m2w_definitions.inc\"\n\n");
				
		target.append("global_settings { ambient_light rgb 1 }\n");
		target.append("light_source{ <100000,150000,-100000> color White fade_power 0 }\n\n");
		
		if (camera != null && projection != null) {
			addCameraDefinition(target, camera, projection);
		}
		
		target.append("sky_sphere {\n");
		target.append("  pigment { Blue_Sky3 }\n");
		target.append("}\n\n");
		
		target.append("difference {\n");
		target.append("  plane { y, -0.001 }\n  ");
		VectorXZ[] boundary = eleData.getBoundaryPolygon().getXZPolygon()
			.getVertexLoop().toArray(new VectorXZ[0]);
		target.appendPrism( -100, 1, boundary);
		target.append("\n");
		target.appendMaterial(Materials.EMPTY_GROUND);
		target.append("\n}\n\n");
		
		target.append("\n\n//\n//Grid\n//\n\n");
				
		addDefinitionsForWorldObjects(target, grid);

		target.append("\n\n//\n//Terrain\n//\n\n");
		
		terrain.renderTo(target);
		
	}

	private static final void addDefinitionsForWorldObjects(
			final POVRayTarget target, MapData grid) {
		
		iterate(grid.getMapElements(), new Operation<MapElement>() {
			@Override public void perform(MapElement e) {
				
				for (WorldObject representation : e.getRepresentations()) {
		        	if (representation instanceof RenderableToPOVRay) {
		        		RenderableToPOVRay r = (RenderableToPOVRay)representation;
						r.renderTo(target);
		        	} else if (representation instanceof RenderableToAllTargets) {
		        		RenderableToAllTargets r = (RenderableToAllTargets)representation;
						r.renderTo(target);
		        	}
				}
				
			}
		});
        
	}
		
	private static final void addCameraDefinition(POVRayTarget target,
			Camera camera, Projection projection) {
		
		target.append("camera {");
		
		if (projection.isOrthographic()) {
			target.append("\n  orthographic");
		}
		
		target.append("\n  location ");
		target.appendVector(camera.getPos());
		
		if (projection.isOrthographic()) {
			
			target.append("\n  right ");
			double width = projection.getVolumeHeight()
				* projection.getAspectRatio();
			target.appendVector(camera.getRight().mult(width).invert()); //invert compensates for left-handed vs. right handed coordinates
			
			target.append("\n  up ");
			VectorXYZ up = camera.getRight().cross(camera.getViewDirection());
			target.appendVector(up.normalize().mult(projection.getVolumeHeight()));
						
			target.append("\n  look_at ");
			target.appendVector(camera.getLookAt());
						
		} else {
			
			target.append("\n  look_at  ");
			target.appendVector(camera.getLookAt());
			
		}
		
		target.append("\n}\n\n");
		
	}

	private static final void addRenderParameters(POVRayTarget target) {
		
		target.append("//chooses the season (possible values: spring, summer, autumn, winter)\n");
		target.append("#declare season=\"summer\";\n");
		target.append("//chooses the time of day (possible values: day, night)\n");
		target.append("#declare time=\"day\";\n\n\n");
				
	}
	
}
